"""vocabularies commands for cmem command line interface."""

from six.moves.urllib.parse import quote

import click

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.vocabularies import (
    get_global_vocabs_cache,
    get_vocabularies,
    install_vocabulary,
    uninstall_vocabulary
)
from cmem.cmempy.workspace import update_global_vocabulary_cache


def _transform_cache_to_table(cache_category, table):
    """Transform a cache category dict to a tabulate table."""
    for item in cache_category:
        uri = item["genericInfo"]["uri"]
        try:
            label = item["genericInfo"]["label"]
        except KeyError:
            label = ""
        row = [uri, "class", label]
        table.append(row)
    return table


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "iri",
    type=click.STRING,
    autocompletion=completion.installed_vocabularies
)
@click.pass_obj
def open_command(app, iri):
    """Open / explore a vocabulary graph in the browser.

    Vocabularies are identified by their graph IRI.
    Installed vocabularies can be listed with the "vocabulary list" command.
    """
    explore_uri = get_cmem_base_uri() + "/explore?graph=" + quote(iri)
    click.launch(explore_uri)
    app.echo_debug(explore_uri)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only vocabulary identifier (IRIs) and no labels or other "
         "meta data. This is useful for piping the ids into other cmemc "
         "commands."
)
@click.option(
    "--filter", "filter_",
    type=click.Choice(
        ["all", "installed", "installable"],
        case_sensitive=True
    ),
    default="installed",
    show_default=True,
    help="Filter list based on status."
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.pass_obj
def list_command(app, id_only, filter_, raw):
    """Output a list of vocabularies.

    Vocabularies are graphs (see 'cmemc graph' command group) which consists
    of class and property descriptions.
    """
    vocabs = get_vocabularies(filter_=filter_)
    if raw:
        app.echo_info_json(vocabs)
    elif id_only:
        for _ in vocabs:
            app.echo_info(_["iri"])
    else:
        table = []
        for _ in vocabs:
            iri = _["iri"]
            try:
                label = _["label"]["title"]
            except (KeyError, TypeError):
                if _["vocabularyLabel"]:
                    label = _["vocabularyLabel"]
                else:
                    label = "[no label given]"
            table.append((iri, label))
        # sort output by label - https://docs.python.org/3/howto/sorting.html
        table = sorted(table, key=lambda k: k[1].lower())
        app.echo_info_table(
            table,
            headers=["Vocabulary Graph IRI", "Label"]
        )


@click.command(cls=CmemcCommand, name="install")
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.installable_vocabularies
)
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Install all vocabularies from the catalog."
)
@click.pass_obj
def install_command(app, iris, all_):
    """Install one or more vocabularies from the catalog.

    Vocabularies are identified by their graph IRI.
    Installable vocabularies can be listed with the
    "vocabulary list --filter installable" command.
    """
    if iris == () and not all_:
        raise ValueError("Either specify at least one vocabulary "
                         + "IRI or use the --all option to install all "
                           "vocabularies from the catalog.")
    if all_:
        iris = (
            [iri["iri"] for iri in get_vocabularies(filter_="installable")]
        )
    count: int = len(iris)
    current: int = 1
    for iri in iris:
        app.echo_info(
            "Install vocabulary {}/{}: {} ... ".format(current, count, iri),
            nl=False)
        install_vocabulary(iri)
        app.echo_success("done")
        current = current + 1


@click.command(cls=CmemcCommand, name="uninstall")
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.installed_vocabularies
)
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Uninstall all installed vocabularies."
)
@click.pass_obj
def uninstall_command(app, iris, all_):
    """Uninstall one or more vocabularies.

    Vocabularies are identified by their graph IRI.
    Already installed vocabularies can be listed with the
    "vocabulary list --filter installed" command.
    """
    if iris == () and not all_:
        raise ValueError("Either specify at least one vocabulary "
                         + "IRI or use the --all option to uninstall all "
                           "installed vocabularies.")
    if all_:
        iris = (
            [iri["iri"] for iri in get_vocabularies(filter_="installed")]
        )
    count: int = len(iris)
    current: int = 1
    for iri in iris:
        app.echo_info(
            "Uninstall vocabulary {}/{}: {} ... ".format(current, count, iri),
            nl=False)
        uninstall_vocabulary(iri)
        app.echo_success("done")
        current = current + 1


@click.command(cls=CmemcCommand, name="update")
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.installed_vocabularies
)
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Update cache for all installed vocabularies."
)
@click.pass_obj
def cache_update_command(app, iris, all_):
    """Reload / updates the data integration cache for a vocabulary."""
    if iris == () and not all_:
        raise ValueError("Either specify at least one vocabulary "
                         + "IRI or use the --all option to update the "
                           "cache for all installed vocabularies.")
    if all_:
        iris = (
            [iri["iri"] for iri in get_vocabularies(filter_="installed")]
        )
    count: int = len(iris)
    current: int = 1
    for iri in iris:
        app.echo_info(
            "Update cache {}/{}: {} ... ".format(current, count, iri),
            nl=False)
        update_global_vocabulary_cache(iri)
        app.echo_success("done")
        current = current + 1


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only vocabulary term identifier (IRIs) and no labels or other "
         "meta data. This is useful for piping the ids into other cmemc "
         "commands."
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.pass_obj
def cache_list_command(app, id_only, raw):
    """Output the content of the global vocabulary cache."""
    cache_ = get_global_vocabs_cache()
    if raw:
        app.echo_info_json(cache_)
    elif id_only:
        for vocab in cache_["vocabularies"]:
            for class_ in vocab["classes"]:
                app.echo_info(class_["genericInfo"]["uri"])
            for property_ in vocab["properties"]:
                app.echo_info(property_["genericInfo"]["uri"])
    else:
        table = []
        for vocab in cache_["vocabularies"]:
            table = _transform_cache_to_table(vocab["classes"], table)
            table = _transform_cache_to_table(vocab["properties"], table)
        app.echo_info_table(
            table,
            headers=["IRI", "Type", "Label"]
        )


@click.group(cls=CmemcGroup)
def cache():
    """List und update the vocabulary cache."""


cache.add_command(cache_update_command)
cache.add_command(cache_list_command)


@click.group(cls=CmemcGroup)
def vocabulary():
    """List, install, uninstall and open vocabularies."""


vocabulary.add_command(open_command)
vocabulary.add_command(list_command)
vocabulary.add_command(install_command)
vocabulary.add_command(uninstall_command)
vocabulary.add_command(cache)



"""The main command line interface."""

import json
import os
import sys
import traceback

import click

import requests.exceptions

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import (
    admin,
    config,
    dataset,
    graph,
    project,
    query,
    vocabulary,
    workflow,
    workspace
)
from cmem.cmemc.cli.context import CONTEXT
from cmem.cmemc.cli.exceptions import (
    InvalidConfiguration
)
from cmem.cmemc.cli.utils import is_completing
from cmem.cmemc.cli.commands import CmemcGroup

try:
    from cmem.cmemc.cli.version import VERSION
except ImportError:
    VERSION = "SNAPSHOT"

PYTHON_VERSION = "{}.{}.{}".format(
    sys.version_info.major,
    sys.version_info.minor,
    sys.version_info.micro
)
PYTHON_EXPECTED = "3.7"
PYTHON_GOT = "{}.{}".format(
    sys.version_info.major,
    sys.version_info.minor
)
if PYTHON_EXPECTED != PYTHON_GOT:
    # test for environment which indicates that we are in completion mode
    if not is_completing():
        CONTEXT.echo_warning(
            "Warning: Your are running cmemc under a non-tested python "
            "environment (expected {}, got {})"
            .format(PYTHON_EXPECTED, PYTHON_GOT)
        )

# set the user-agent environment for the http request headers
os.environ["CMEM_USER_AGENT"] = "cmemc/{} (Python {})"\
                                .format(VERSION, PYTHON_VERSION)

# https://github.com/pallets/click/blob/master/examples/complex/complex/cli.py
CONTEXT_SETTINGS = dict(
    auto_envvar_prefix='CMEMC',
    help_option_names=['-h', '--help']
)


@click.group(cls=CmemcGroup, context_settings=CONTEXT_SETTINGS)
@click.option(
    '-c', '--connection',
    type=click.STRING,
    autocompletion=completion.connections,
    help='Use a specific connection from the config file.'
)
@click.option(
    '--config-file',
    autocompletion=completion.ini_files,
    type=click.Path(
        readable=True,
        allow_dash=False,
        dir_okay=False
    ),
    default=CONTEXT.config_file_default,
    show_default=True,
    help='Use this config file instead of the default one.'
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Suppress any non-error info messages.'
)
@click.option(
    '-d', '--debug',
    is_flag=True,
    help='Output debug messages and stack traces after errors.'
)
@click.version_option(
    version=VERSION,
    message="%(prog)s, version %(version)s, "
            "running under python {}".format(PYTHON_VERSION)
)
@click.pass_context
def cli(ctx, debug, quiet, config_file, connection):  # noqa: D403
    """eccenca Corporate Memory Control (cmemc).

    cmemc is the eccenca Corporate Memory Command Line Interface (CLI).

    Available commands are grouped by affecting resource type (such as graph,
    project and query).
    Each command and group has a separate --help screen for detailed
    documentation.
    In order to see possible commands in a group, simply
    execute the group command without further parameter (e.g. cmemc project).

    If your terminal supports colors, these coloring rules are applied:
    Groups are colored in white; Commands which change data are colored in
    red; all other commands as well as options are colored in green.

    Please also have a look at the cmemc online documentation:

                        https://eccenca.com/go/cmemc

    cmemc is © 2021 eccenca GmbH, licensed under the Apache License 2.0.
    """
    ctx.obj = CONTEXT
    ctx.obj.set_quiet(quiet)
    ctx.obj.set_debug(debug)
    ctx.obj.set_config_file(config_file)
    try:
        ctx.obj.set_connection(connection)
    except InvalidConfiguration as error:
        # if config is broken still allow for "config edit"
        # means: do not forward this exception if "config edit"
        if " ".join(sys.argv).find("config edit") == -1:
            raise error


cli.add_command(config.config)
cli.add_command(workspace.workspace)
cli.add_command(graph.graph)
cli.add_command(query.query)
cli.add_command(project.project)
cli.add_command(vocabulary.vocabulary)
cli.add_command(workflow.workflow)
cli.add_command(dataset.dataset)
cli.add_command(admin.admin)


def main():
    """Start the command line interface."""
    try:
        cli()  # pylint: disable=no-value-for-parameter
    except (
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError,
            ValueError,
            IOError,
            NotImplementedError,
            KeyError
    ) as error:
        if is_completing():
            # if currently autocompleting -> silently die with exit 1
            sys.exit(1)
        CONTEXT.check_versions()
        CONTEXT.echo_error(type(error).__name__ + ": " + str(error))
        CONTEXT.echo_debug(traceback.format_exc())
        # exceptions with response is HTTPError
        try:
            # try to load Problem Details for HTTP API JSON
            details = json.loads(error.response.text)
            error_message = type(error).__name__ + ": "
            if 'title' in details:
                error_message += details["title"] + ": "
            if 'detail' in details:
                error_message += details["detail"]
            CONTEXT.echo_error(error_message)
        except (AttributeError, ValueError):
            # is not json or any other issue, output plain response text
            pass
        sys.exit(1)



"""DataIntegration project commands for the cmem command line interface."""

import os
from pathlib import Path

import click

from jinja2 import Template

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmempy.config import get_di_api_endpoint
from cmem.cmempy.plugins.marshalling import (
    get_extension_by_plugin,
    get_plugin_by_extension
)
from cmem.cmempy.workspace.projects.export_ import export_project
from cmem.cmempy.workspace.projects.import_ import import_project
from cmem.cmempy.workspace.projects.project import (
    delete_project,
    get_projects,
    make_new_project
)


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "project_ids",
    nargs=-1,
    required=True,
    type=click.STRING,
    autocompletion=completion.project_ids
)
@click.pass_obj
def open_command(app, project_ids):
    """Open projects in the browser.

    With this command, you can open a project in the workspace in
    your browser to change them.

    The command accepts multiple projects IDs which results in
    opening multiple browser tabs.
    """
    projects = get_projects()
    for _ in project_ids:
        if _ not in (p["name"] for p in projects):
            raise ValueError("Project '{}' not found.".format(_))
        open_project_uri = "{}/workbench/projects/{}".format(
            get_di_api_endpoint(),
            _
        )
        app.echo_debug("Open {}: {}".format(_, open_project_uri))
        click.launch(open_project_uri)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only project identifier and no labels or other "
         "meta data. This is useful for piping the IDs into other commands."
)
@click.pass_obj
def list_command(app, raw, id_only):
    """List available projects.

    Outputs a list of project IDs which can be used as reference for
    the project create, delete, export and import commands.
    """
    projects = get_projects()
    if raw:
        app.echo_info_json(projects)
        return
    if id_only:
        for _ in sorted(projects, key=lambda k: k["name"].lower()):
            app.echo_result(_["name"])
        return
    # output a user table
    table = []
    for _ in projects:
        row = [
            _["name"],
            _["metaData"]["label"],
        ]
        table.append(row)
    # sort output by label - https://docs.python.org/3/howto/sorting.html
    table = sorted(table, key=lambda k: k[1].lower())
    app.echo_info_table(
        table,
        headers=["Project ID", "Label"]
    )


@click.command(cls=CmemcCommand, name="delete")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Delete all projects. "
         "This is a dangerous option, so use it with care.",
)
@click.argument(
    "project_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.project_ids
)
@click.pass_obj
def delete_command(app, all_, project_ids):
    """Delete project(s).

    This deletes existing data integration projects from Corporate Memory.
    Projects will be deleted without prompting!

    Example: cmemc project delete my_project

    Projects can be listed by using the 'cmemc project list' command.
    """
    if project_ids == () and not all_:
        raise ValueError("Either specify at least one project ID "
                         + "or use the --all option to delete all projects.")
    if all_:
        # in case --all is given, a list of project is fetched
        project_ids = (
            [currentProjects["name"] for currentProjects in get_projects()]
        )
    count = len(project_ids)
    current = 1
    for project_id in project_ids:
        app.echo_info("Delete project {}/{}: {} ... "
                      .format(current, count, project_id), nl=False)
        # always fetch the current list of projects
        if project_id in ([_["name"] for _ in get_projects()]):
            delete_project(project_id)
            app.echo_success("done")
        else:
            app.echo_error("not found")
        current = current + 1


@click.command(cls=CmemcCommand, name="create")
@click.argument(
    "project_ids",
    nargs=-1,
    required=True,
    type=click.STRING
)
@click.pass_obj
def create_command(app, project_ids):
    """Create empty new project(s).

    This creates one or more new projects.
    Existing projects will not be overwritten.

    Example: cmemc project create my_project

    Projects can be listed by using the 'cmemc project list' command.
    """
    count = len(project_ids)
    current = 1
    for project_id in project_ids:
        app.echo_info("Create new project {}/{}: {} ... "
                      .format(current, count, project_id), nl=False)
        app.echo_debug(get_projects())
        # always fetch the current list of projects
        if project_id not in ([_["name"] for _ in get_projects()]):
            make_new_project(project_id)
            app.echo_success("done")
        else:
            app.echo_error("already there")
        current = current + 1


# pylint: disable=too-many-arguments
@click.command(cls=CmemcCommand, name="export")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Export all projects.",
)
@click.option(
    "-o", "--overwrite",
    is_flag=True,
    help="Overwrite existing files. This is a dangerous option, "
         "so use it with care.",
)
@click.option(
    "--output-dir",
    default=".",
    show_default=True,
    type=click.Path(
        writable=True,
        file_okay=False
    ),
    # autocompletion=directories,
    help="Directory, where the project files will be created. "
         "This directory will be created as well it does not exists."
)
@click.option(
    "--type", "marshalling_plugin",
    default="xmlZip",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.marshalling_plugins,
    help="Type of the exported project file(s)."
)
@click.option(
    "--filename-template", "-t", "template",
    default="{{date}}-{{connection}}-{{id}}.project",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.project_export_templates,
    help="Template for the export file name(s). "
         "Possible placeholders are (Jinja2): "
         "{{id}} (the project ID), "
         "{{connection}} (from the --connection option) and "
         "{{date}} (the current date as YYYY-MM-DD). "
         "The file suffix will be appended. "
         "Needed directories will be created."
)
@click.argument(
    "project_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.project_ids
)
@click.pass_obj
def export_command(
        app, all_, project_ids, overwrite, marshalling_plugin,
        template, output_dir):
    """Export project(s) to file(s).

    Projects can be exported with different export formats.
    The default type is a zip archive which includes meta data as well
    as dataset resources.
    If more than one project is exported, a file is created for each project.
    By default, these files are created in the current directory and with
    a descriptive name (see --template option default).

    Example: cmemc project export my_project

    Available projects can be listed by using the 'cmemc project list' command.

    You can use the template string to create subdirectories as well:
    cmemc config list | parallel -I% cmemc -c % project export --all
    -t "dump/{{connection}}/{{date}}-{{id}}.project"
    """
    # pylint: disable=too-many-arguments
    if project_ids == () and not all_:
        raise ValueError("Either specify at least one project ID "
                         + "or use the --all option to export all projects.")
    if all_:
        # in case --all is given, a list of project is fetched
        project_ids = ([_["name"] for _ in get_projects()])

    template_data = app.get_template_data()

    count = len(project_ids)
    current = 1
    for project_id in project_ids:
        # prepare the template data and create the actual file incl. suffix
        template_data.update(id=project_id)
        file_name = Template(template).render(template_data)\
            + "." + get_extension_by_plugin(marshalling_plugin)
        # join with given output directory and normalize full path
        export_path = os.path.normpath(os.path.join(output_dir, file_name))
        app.echo_info("Export project {}/{}: {} to {} ... "
                      .format(current, count, project_id, export_path),
                      nl=False)
        if os.path.exists(export_path) and overwrite is not True:
            app.echo_error("file exists")
        else:
            # always fetch the current list of projects
            if project_id in ([_["name"] for _ in get_projects()]):
                export_data = export_project(project_id, marshalling_plugin)
                # output directory is created lazy
                Path(
                    os.path.dirname(export_path)
                ).mkdir(parents=True, exist_ok=True)
                # do the export
                with open(export_path, "wb") as export_file:
                    export_file.write(export_data)
                app.echo_success("done")
            else:
                app.echo_error("project not found")
        current = current + 1


@click.command(cls=CmemcCommand, name="import")
@click.argument(
    "file",
    autocompletion=completion.project_files,
    type=click.Path(
        allow_dash=False,
        dir_okay=False,
        readable=True
    )
)
@click.argument(
    "project_id",
    type=click.STRING,
    autocompletion=completion.project_ids
)
@click.pass_obj
def import_command(app, file, project_id):
    """Import a project from a file.

    Example: cmemc project import my_project.zip my_project
    """
    file_extension = os.path.splitext(file)[1][1:]
    plugin_id = get_plugin_by_extension(file_extension)
    app.echo_info("Import file {} to project {} ... "
                  .format(file, project_id),
                  nl=False)
    if project_id not in ([_["name"] for _ in get_projects()]):
        import_project(project_id, file, plugin_id)
        app.echo_success("done")
    else:
        app.echo_error("already there")


@click.group(cls=CmemcGroup)
def project():
    """List, import, export, create or delete projects.

    Projects are identified by an PROJECTID. The get a list of existing
    projects, execute the list command or use tab-completion.
    """


project.add_command(open_command)
project.add_command(list_command)
project.add_command(export_command)
project.add_command(import_command)
project.add_command(delete_command)
project.add_command(create_command)



"""configuration commands for cmem command line interface."""
import click

from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup


@click.command(cls=CmemcCommand, name="list")
@click.pass_obj
def list_command(app):
    """List configured connections.

    This command lists all configured
    connections from the currently used config file.

    The connection identifier can be used with the --connection option
    in order to use a specific Corporate Memory instance.

    In order to apply commands on more than one instance, you need to use
    typical unix gear such as xargs or parallel:

    cmemc config list | xargs -I % sh -c 'cmemc -c % admin status'

    cmemc config list | parallel --jobs 5 cmemc -c {} admin status
    """
    for section_string in sorted(app.config, key=str.casefold):
        if section_string != "DEFAULT":
            app.echo_result(section_string)


@click.command(cls=CmemcCommand, name="edit")
@click.pass_obj
def edit_command(app):
    """Edit the user-scope configuration file."""
    app.echo_info("Open editor for config file " + app.config_file)
    click.edit(filename=app.config_file)


@click.group(cls=CmemcGroup)
def config():
    u"""List and edit configurations.

    Configurations are identified by the section identifier in the
    config file. Each configuration represent a Corporate Memory deployment
    with its specific access method as well as credentials.

    A minimal configuration which uses client credentials has the following
    entries:

    \b
    [example.org]
    CMEM_BASE_URI=https://cmem.example.org/
    OAUTH_GRANT_TYPE=client_credentials
    OAUTH_CLIENT_ID=cmem-service-account
    OAUTH_CLIENT_SECRET=my-secret-account-pass

    Note that OAUTH_GRANT_TYPE can be either client_credentials, password or
    prefetched_token.

    In addition to that, the following config parameters can be used as well:

    \b
    SSL_VERIFY=False    - for ignoring certificate issues (not recommended)
    DP_API_ENDPOINT=URL - to point to a non-standard DataPlatform location
    DI_API_ENDPOINT=URL - to point to a non-standard DataIntegration location
    OAUTH_TOKEN_URI=URL - to point to an external IdentityProvider location
    OAUTH_USER=username - only if OAUTH_GRANT_TYPE=password
    OAUTH_PASSWORD=password - only if OAUTH_GRANT_TYPE=password
    OAUTH_ACCESS_TOKEN=token - only if OAUTH_GRANT_TYPE=prefetched_token

    In order to request passwords on start, you can use the following
    parameter instead the PASSWORD parameter: OAUTH_PASSWORD_ENTRY=True
    OAUTH_CLIENT_SECRET_ENTRY=True.
    """


config.add_command(list_command)
config.add_command(edit_command)
